from ._scholarly import _Scholarly
from ._proxy_generator import ProxyGenerator
from .data_types import Author, Publication
from ._navigator import DOSException, MaxTriesExceededException
scholarly = _Scholarly()
